package com.ekexiu.portal.deploy;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.security.KeyStore.Entry;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import com.google.gson.Gson;
import com.google.gson.reflect.TypeToken;

import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import freemarker.template.TemplateExceptionHandler;

public class Main {
	public static Charset UTF8 = Charset.forName("UTF-8");

	private File source;
	private File dest;
	private String model = "dev";
	private byte[] copyBuf = new byte[4096];

	private boolean debug = false;

	private Configuration cfg;

	private Map<String, Object> setting;

	private LinkedList<String> fileMatchers = new LinkedList<String>();
	private LinkedList<String> ignoreFileMatchers = new LinkedList<String>();
	private AntPathMatcher pathMatcher = new AntPathMatcher();
	private Gson gson = new Gson();

	private LinkedList<String> currentFileName = new LinkedList<String>();

	public File getSource() {
		return source;
	}

	public void setSource(File source) {
		this.source = source;
		log("set source dir["+source.getAbsolutePath()+"]");
	}

	public File getDest() {
		return dest;
	}

	public void setDest(File dest) {
		log("set dest dir["+dest.getAbsolutePath()+"]");
		this.dest = dest;
	}

	public String getModel() {
		return model;
	}

	public void setModel(String model) {
		this.model = model;
		log("set model["+model+"]");
	}

	public Configuration getCfg() {
		return cfg;
	}

	public void setCfg(Configuration cfg) {
		this.cfg = cfg;
	}

	public void log(String s) {
		if (debug) {
			System.out.println(s);
		}
	}

	public static void main(String[] args) {
		if (args.length < 3) {
			showUseage();
			return;
		}
		File s = new File(args[0]);
		File d = new File(args[1]);
		String dm = args.length > 2 ? (args[2]) : "dev";
		dm = dm.trim();
		if (dm.length() == 0)
			dm = "dev";
		if (!s.exists()) {
			System.out.println("source path [" + s.getAbsolutePath() + "] not exists");
			return;
		}
		if (s.isFile()) {
			System.out.println("source path [" + s.getAbsolutePath() + "] is a file");
			return;
		}
		try {
			Main.checkDir(d);
			Main command = new Main();
			command.debug = args.length > 3;
			command.setDest(d);
			command.setSource(s);
			command.setModel(dm);
			command.initSetting();
			command.initConfig();
			command.processDir(command.getSource());
			System.out.println("deploy success");
		} catch (Throwable th) {
			th.printStackTrace(System.out);
		}
	}

	public static void showUseage() {
		System.out.println("usage:  deploy  SOURCE DEST  DEPLOY_MODEL");
		System.out.println("            SOURCE  file path ");
		System.out.println("            DEST       file path ");
		System.out.println("            DEPLOY_MODEL     like  dev test  pub... ");
	}

	public String getCurrentFileName() {
		StringBuilder sb = new StringBuilder();
		for (ListIterator<String> it = this.currentFileName.listIterator(); it.hasNext();) {
			sb.append(it.next());
		}
		return sb.toString();
	}

	public void initConfig() throws IOException {
		this.cfg = new Configuration(Configuration.VERSION_2_3_26);
		cfg.setDirectoryForTemplateLoading(this.source);
		cfg.setTemplateExceptionHandler(TemplateExceptionHandler.RETHROW_HANDLER);
	}

	public void processDir(File dir) throws IOException, TemplateException {
		for (String fn : dir.list()) {
			if(debug){
				System.out.println("list dir["+dir.getAbsolutePath()+"]===>child["+fn+"]");
			}
			if (fn.equals(".") || fn.equals("..")) {
				continue;
			} else {
				this.currentFileName.addLast("/" + fn);
				try {
					File file = new File(dir, fn);
					if (file.isFile()) {
						processFile(file);
					} else {
						processDir(file);
					}
				} finally {
					this.currentFileName.removeLast();
				}
			}
		}
	}

	public boolean isMatchHandler(String fn) {
		for (String p : fileMatchers) {
			if (this.pathMatcher.match(p, fn))
				return true;
		}
		return false;
	}

	public boolean isMatchIgnore(String fn) {
		for (String p : this.ignoreFileMatchers) {
			if (this.pathMatcher.match(p, fn))
				return true;
		}
		return false;
	}

	public void processFile(File file) throws IOException, TemplateException {
		String fullName = this.getCurrentFileName();
		String name = this.currentFileName.getLast().substring(1);
		File destFile = null;

		if (name.endsWith(".ftl")) {
			name = name.substring(0, name.length() - 4);
			if (name.indexOf('.') <= 0) {
				if(debug){
					System.out.println("file["+fullName+"] is only ftl    don't copy and don't  process");
				}
				return;
			}
			destFile = new File(this.dest, fullName.substring(1, fullName.length() - 4));
		}
		if (destFile == null) {
			if (this.isMatchHandler(fullName)) {
				destFile = new File(this.dest, fullName.substring(1));
			}
		}

		if (destFile == null) {
			if (isMatchIgnore(fullName)) {
				if(debug){
					System.out.println("file["+fullName+"] is ignored");
				}
				return;
			}
			destFile = new File(this.dest, fullName.substring(1));
			if(debug){
				System.out.println("file["+fullName+"] will do  copy to file:"+destFile.getAbsolutePath());
			}
			checkDir(destFile.getParentFile());
			copy(file, destFile);
		} else {
			if(debug){
				System.out.println("file["+fullName+"] will do  process to file:"+destFile.getAbsolutePath());
			}
			checkDir(destFile.getParentFile());
			genFile(fullName, destFile);
		}

	}

	private void copy(File src, File destFile) throws IOException {
		if (destFile.exists()) {
			if (!destFile.delete()) {
				throw new IOException("delete file[" + destFile.getAbsolutePath() + "] error");
			}
		}
		InputStream in = new FileInputStream(src);
		try {
			OutputStream out = new FileOutputStream(destFile);
			try {
				int len = 0;
				while ((len = in.read(copyBuf)) != -1) {
					if (len > 0) {
						out.write(copyBuf, 0, len);
					}
				}
				out.flush();
			} finally {
				out.close();
			}
		} finally {
			in.close();
		}

	}

	private void genFile(String name, File destFile) throws IOException, TemplateException {
		Template template = this.cfg.getTemplate(name, "UTF-8");
		this.setting.put("templateFileName", name);
		if (destFile.exists()) {
			if (!destFile.delete()) {
				throw new IOException("delete file[" + destFile.getAbsolutePath() + "] error");
			}
		}
		OutputStream out = new FileOutputStream(destFile);
		try {
			template.process(this.setting, new OutputStreamWriter(out, UTF8));
			out.flush();
		} finally {
			out.close();
		}
	}

	private void merge(Map<String, Object> dest, Map<String, Object> src) {
		for (Map.Entry<String, Object> entry : src.entrySet()) {
			String key = entry.getKey();
			Object val = entry.getValue();
			Object dval = dest.get(key);
			if (dval == null) {
				dest.put(key, val);
			} else {
				if (dval instanceof Number) {
					dest.put(key, val);
				} else if (dval instanceof Boolean) {
					dest.put(key, val);
				}
				if (dval instanceof CharSequence) {
					dest.put(key, val);
				} else if (dval instanceof List) {
					if (val instanceof List) {
						merge((List<Object>) dval, (List<Object>) val);
					} else {
						((List<Object>) dval).add(val);
					}
				} else if (dval instanceof Map) {
					if (val instanceof Map) {
						merge((Map<String, Object>) dval, (Map<String, Object>) val);
					}
				}
			}
		}
	}

	private void merge(List<Object> dest, List<Object> src) {
		for (Object obj : src) {
			if (!dest.contains(obj)) {
				dest.add(obj);
			}
		}
	}

	private Map<String, Object> readJson(File file, boolean raise) throws IOException {
		if (!file.exists()) {
			if (raise)
				throw new IOException("file " + file.getAbsolutePath() + " not exists");
			return null;
		}
		if (file.isDirectory()) {
			if (raise)
				throw new IOException("file " + file.getAbsolutePath() + " not exists");
			return null;
		}
		InputStream in = new FileInputStream(file);
		try {
			return gson.fromJson(new InputStreamReader(in, "UTF-8"), new TypeToken<Map<String, Object>>() {
			}.getType());
		} finally {
			in.close();
		}
	}

	private void initSetting() throws IOException {
		setting = readJson(new File(source, "build_cfg/ekexiu_build.json"), true);
		if(debug){
			System.out.println(gson.toJson(setting));
		}
		Map<String, Object> appendSetting = readJson(new File(source, "build_cfg/ekexiu_build_" + model + ".json"), false);
		if(debug){
			System.out.println(gson.toJson(appendSetting));
		}
		merge(setting, appendSetting);
		setting.put("deploy", model);
		if(debug){
			System.out.println(gson.toJson(setting));
		}
		Object fs = setting.get("handleFiles");
		if (fs.getClass().equals(String.class)) {
			this.fileMatchers.add((String) fs);
		} else if (fs instanceof List) {
			for (Object item : (List) fs) {
				if (item != null && item.getClass().equals(String.class)) {
					this.fileMatchers.add((String) item);
				}
			}
		}
		Object iis = setting.get("ignoreFiles");
		if (iis.getClass().equals(String.class)) {
			this.ignoreFileMatchers.add((String) iis);
		} else if (iis instanceof List) {
			for (Object item : (List) iis) {
				if (item != null && item.getClass().equals(String.class)) {
					this.ignoreFileMatchers.add((String) item);
				}
			}
		}
		if(debug){
			System.out.println(gson.toJson(setting));
		}
	}

	private static void checkDir(File dir) throws IOException {
		if (!dir.exists()) {
			if (!dir.mkdirs()) {
				throw new IOException("mkdir [" + dir.getAbsolutePath() + "] error");
			}
		}
		if (dir.isFile()) {
			throw new IOException("dir [" + dir.getAbsolutePath() + "] is file");
		}
	}
}
